/*
**  NGPT - Next Generation POSIX Threading
**  Copyright (c) 2001 IBM Corporation <babt@us.ibm.com>
**  Portions Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
**
**  This file is part of NGPT, a non-preemptive thread scheduling
**  library which can be found at http://www.ibm.com/developer.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
**  USA.
**
**  pth_tcb.c: Pth thread control block handling
*/
                             /* Patient: Doctor, it hurts when I do this!
                                Doctor: Well, then don't do it. */
#include "pth_p.h"

#if cpp

#define PTH_TCB_NAMELEN 40
typedef pid_t	pth_native_t;	/*ibm*/

    /* thread control block */
struct pth_st {
    /* machine context - LEAVE mctx AS FIRST MEMBER OF THIS STRUCTURE                   */ /*ibm*/
    pth_mctx_t     mctx;                 /* last saved machine state of thread          */
    pth_native_t   boundnative;		 /* Native thread this tcb is bound to or NULL  */ /*ibm*/
    pth_native_t   lastrannative;        /* Last native thread this tcb ran on          */ /*ibm*/
    char          *stack;                /* pointer to thread stack                     */
    char	  *true_stack;		 /* pointer to the true stack                   */
#if PTH_NEED_SEPARATE_REGISTER_STACK
    char	  *regstack;		 /* pointer to register stack on IA64           */ /*ibm*/
#endif
    unsigned int   stacksize;            /* size of thread stack                        */
    long          *stackguard;           /* stack overflow guard                        */
    int            stackloan;            /* stack type                                  */
    void        *(*start_func)(void *);  /* start routine                               */
    void          *start_arg;            /* start argument                              */

    /* priority queue handling */
    pth_t          q_next;               /* next thread in pool                         */
    pth_t          q_prev;               /* previous thread in pool                     */
    int            q_prio;               /* (relative) priority of thread when queued   */

    /* standard thread control block ingredients */
    int            prio;                 /* base priority of thread                     */
    char           name[PTH_TCB_NAMELEN];/* name of thread (mainly for debugging)       */
    pth_state_t    state;                /* current state indicator for thread          */

    /* timing                                                                           */
    pth_time_t     spawned;              /* time point at which thread was spawned      */
    pth_time_t     lastran;              /* time point at which thread was last running */
    pth_time_t     running;              /* time range the thread was already running   */

    /* event handling */
    pth_event_t    events;               /* events the tread is waiting for             */

    /* per-thread signal handling */
    sigset_t       sigpending;           /* set    of pending signals                   */
    int            sigpendcnt;           /* number of pending signals                   */

    /* thread joining */
    int            joinable;             /* whether thread is joinable                  */
    void          *join_arg;             /* joining argument                            */

    /* per-thread specific storage */
    const void   **data_value;           /* thread specific  values                     */
    int            data_count;           /* number of stored values                     */

    /* cancellation support */
    int            cancelreq;            /* cancellation request is pending             */
    unsigned int   cancelstate;          /* cancellation state of thread                */
    pth_cleanup_t *cleanups;             /* stack of thread cleanup handlers            */

    /* mutex ring */
    pth_ring_t     mutexring;            /* ring of aquired mutex structures            */

    /* alignment flag */
    int		   ptrfixed;		 /* 1 if ptr was adjusted, 0 otherwise          */ /*ibm*/

    /* lock */
    pth_qlock_t	   lock;		 /* used to lock the thread if needed           */
};

#endif /* cpp */

intern const char *pth_state_names[] = {
    "scheduler", "new", "ready", "running", "waiting", "dead"
};

#if defined(MINSIGSTKSZ) && !defined(SIGSTKSZ)
#define SIGSTKSZ MINSIGSTKSZ
#endif
#if !defined(SIGSTKSZ)
#define SIGSTKSZ 8192
#endif

/* allocate a thread control block */
intern pth_t pth_tcb_alloc(unsigned int stacksize, void *stackaddr)
{
    pth_t t;
    char *stack = NULL;

    if (stacksize > 0 && stacksize < SIGSTKSZ)
        stacksize = SIGSTKSZ;
    /*
     * Some platforms (IA64) require 16 byte alignment for the
     * jmpbuf.  The following code will force the tcb to be
     * aligned on a 16 byte boundary.  Worst case on a platform
     * that doesn't require it, we waste 8 bytes.
     */
    
    t = mmap(NULL, sizeof(struct pth_st)+8,
	     PROT_READ | PROT_WRITE | PROT_EXEC,
	     MAP_PRIVATE | MAP_ANON,
	     -1, 0);
    if (t == MAP_FAILED)
	return(NULL);

    if ((long)t % 16L != 0L) {                                          /*ibm*/
        t = (pth_t)((char *)((char *)t + 8));                           /*ibm*/
	t->ptrfixed = 1;						/*ibm*/
    } else								/*ibm*/
	t->ptrfixed = 0;						/*ibm*/
                                                                                            
    t->stacksize  = stacksize;
    t->stack      = NULL;
    t->stackguard = NULL;
    t->stackloan  = (stackaddr != NULL ? TRUE : FALSE);
    t->lock.count = 0;
    t->lock.owner = 0;
    t->lock.lock  = 0;
    t->boundnative   = 0; /*ibm*/
    t->lastrannative = 0; /*ibm*/
    if (stacksize > 0) { /* stacksize == 0 means "main" thread */
        if (stackaddr != NULL) {
            t->stack = (char *)(stackaddr);
	    t->true_stack = t->stack;
	} else {
#if PTH_NEED_SEPARATE_REGISTER_STACK > 0
	    size_t  pagesize = getpagesize();				/*ibm*/
	    size_t  npages = stacksize / pagesize;			/*ibm*/

	    /* Both ends of the stack need to page aligned */		/*ibm*/
	    t->stacksize = (stacksize % pagesize == 0)			/*ibm*/
			 ? npages * pagesize				/*ibm*/
			 : (++npages) * pagesize;			/*ibm*/
#endif
	    stack = mmap(NULL, (t->stacksize+8),
			    PROT_READ | PROT_WRITE | PROT_EXEC,
			    MAP_PRIVATE | MAP_ANON,
			    -1, 0);
	    if (stack == MAP_FAILED) {
		errno_shield { munmap(t, sizeof(struct pth_st)+8); }
		return(NULL);
	    }

	    t->true_stack = stack;

	    /* Align it if necessary... */
	    if ((long)stack % 16L != 0L)
		stack = ((char *)(stack + 8));
	    t->stack = stack;

        }
#if PTH_NEED_SEPARATE_REGISTER_STACK > 0
	/*									      ibm 
	 * On IA64 the stack will always grow down.  The register stack will          ibm
	 * grow up towards the main stack.                                            ibm
	 */
	 
	/* Guard is at the lowest address, alignment is guaranteed.        */	    /*ibm*/
	/* Note: this guard address is actually in the middle.  Corruption */	    /*ibm*/
	/* could mean that either the main stack or the the register stack */       /*ibm*/
	/* has overflowed.                                                 */       /*ibm*/
	t->stackguard = (long *)((char *)t->stack + (t->stacksize / 2));	    /*ibm*/

	/* Register stack is at the highest address, alignment guaranteed. */	    /*ibm*/
	t->regstack = (char *)(t->stack + t->stacksize);			    /*ibm*/
#else
#if PTH_STACKGROWTH < 0
        /* guard is at lowest address (alignment is guarrantied) */
        t->stackguard = (long *)((long)t->stack); /* double cast to avoid alignment warning */
#else
        /* guard is at highest address (be careful with alignment) */
        t->stackguard = (long *)(t->stack+(((stacksize/sizeof(long))-1)*sizeof(long)));
#endif
#endif
        *t->stackguard = 0xDEAD;
    } else
	t->true_stack = 0;
    return t;
}

/* free a thread control block */
intern void pth_tcb_free(pth_t t)
{

    int i = 0;
    
    pth_debug2("pth_tcb_free: free called for 0x%lx", t);
    if (t == NULL)
        return;
    pth_acquire_lock(&(t->lock));

    /* Cleanup the stack. */
    if (t->true_stack != NULL && !t->stackloan)
        munmap(t->true_stack, t->stacksize+8);

    /* Run the specific data destructors */
    if (t->data_value != NULL)
	pth_key_destroydata(t);

    /* finish the cleanups... */
    if (t->cleanups != NULL)
        pth_cleanup_popall(t, FALSE);

    /*begin ibm*/
    if (t->boundnative != 0) {
	/*clean up bound thread*/;
    }
    
    pth_release_lock(&(t->lock));

    /*
     * The final step is to remove ourselves from any other native
     * thread that happens to be waiting on a timer event for this
     * thread...
     */
    while (pth_native_list[i].is_used) {
	pth_descr_t descr = &pth_native_list[i];
	if (descr->nexttimer_thread == t && descr->tid != gettid())
	    descr->nexttimer_thread = NULL;
	i++;
    }
    
     
    /* 
     * If we adjusted the pointer previously to align on 16 byte boundary, 
     * we need to put it back the proper way now.
     */
    if (t->ptrfixed)
	t = (pth_t)((char *)((char *)t - 8));
    /*ibm end*/

    munmap(t, sizeof(struct pth_st)+8);
    return;
}

